import datetime
import inspect

from tornado import gen
from tornado.platform.asyncio import to_tornado_future

from event_mq import datetime_help, registry


def cls_str_of_meth(meth):
    mod = inspect.getmodule(meth)
    cls = meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
    return '{0}.{1}'.format(mod.__name__, cls)


def cls_str_of_obj(obj):
    return '{0}.{1}'.format(obj.__class__.__module__, obj.__class__.__name__)


def cls_str_of_cls(cls):
    return '{0}.{1}'.format(cls.__module__, cls.__name__)


def meth_str(meth):
    return '{0}.{1}'.format(meth.__module__, meth.__qualname__)


def pre_serialize(data):
    if isinstance(data, list):
        data = [pre_serialize(item) for item in data]

    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = pre_serialize(value)

    if isinstance(data, datetime.datetime):
        data = datetime_help.get_time_str(data)

    if isinstance(data, datetime.date):
        data = datetime_help.get_date_str(data)

    return data


def pre_deserialize(data):
    if isinstance(data, list):
        data = [pre_deserialize(item) for item in data]

    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = pre_deserialize(value)
    if isinstance(data, str):
        value = datetime_help.parse_datetime(data)
        if not value:
            value = datetime_help.parse_date(data)

        data = value or data

    return data


def run_async(func, *args, **kwargs):
    if not inspect.iscoroutinefunction(func) and not gen.is_coroutine_function(func):
        fu = registry.executor_pool().submit(func, *args, **kwargs)
        fu = to_tornado_future(fu)
    else:
        fu = func(*args, **kwargs)
    return gen.convert_yielded(fu)


def import_from_str(cls_path):
    module_name, class_name = cls_path.rsplit('.', 1)
    module_meta = __import__(module_name, globals(), locals(), [class_name])
    class_meta = getattr(module_meta, class_name)
    return class_meta



# coding: utf-8
import json
import time

from tornado import gen

from event_mq import datetime_help, util, registry


class DomainEvent:
    def __init__(self, event_name, **kwargs):
        # self.version = None
        # self.aggregator_id = None
        # self.command_id = None
        self.occur_on = datetime_help.get_utc_time()
        self.event_name = event_name
        self.__dict__.update(kwargs)


def _dumps(event):
    data = {}
    for k, v in event.__dict__.items():
        if k.startswith('__') and k.endswith('__'):
            continue
        data[k] = v
    data = util.pre_serialize(data)
    return json.dumps(data)


def _loads(data):
    data = json.loads(data)
    data = util.pre_deserialize(data)
    return DomainEvent(**data)

async def publish_async(event, mq_server='default'):
    server = registry.get_mq_server(mq_server)
    if not server:
        raise Exception('cannot find named "{0}" mq_server'.format(mq_server))
    return await server.publish_async(event.event_name, _dumps(event))


def publish(event, mq_server='default'):
    future = gen.convert_yielded(publish_async(event, mq_server))
    while not future.done():
        time.sleep(0.01)
    return future.result()


def subscribe(event_name, subscriber, mq_server='default'):
    server = registry.get_mq_server(mq_server)
    if not server:
        raise Exception('cannot find named "{0}" mq_server'.format(mq_server))
    queue = util.meth_str(subscriber)
    server.add_consumer(queue, event_name, subscriber)



# coding: utf8
from concurrent.futures import ThreadPoolExecutor

from tornado import ioloop

from event_mq import settings


class Registry:
    _mq_servers = {}
    _io_loop = ioloop.IOLoop.instance()
    _executor_pool = ThreadPoolExecutor(settings.get('EXECUTOR_MAX_WORKERS'))

    @classmethod
    def _add_mq_server(cls, name, mq_server):
        cls._mq_servers[name] = mq_server


def get_mq_server(name):
    return Registry._mq_servers.get(name, None)


def io_loop():
    return Registry._io_loop


def executor_pool():
    return Registry._executor_pool



# coding: utf-8
import logging
import traceback

import pika
from pika import adapters
from tornado import concurrent, gen

from event_mq import util, event, registry, settings

LOGGER = logging.getLogger(__name__)


def _get_channel(connection):
    f = concurrent.Future()

    def on_open(channel):
        f.set_result(channel)

    if not connection.is_open:
        f.set_exception(Exception('connection is closed, cannot get channel'))

    connection.channel(on_open_callback=on_open)
    return f


class Consumer:
    def __init__(self, exchange, queue, routing_key, handler, io_loop=None):
        self._exchange = exchange
        self._queue = queue
        self._routing_key = routing_key
        self._handler = handler
        self.running = False
        self._connection = None
        self._io_loop = io_loop or registry.io_loop()

    def __on_message(self, channel, deliver, properties, body):
        encoding = properties.content_encoding or 'utf-8'
        body = body.decode(encoding)
        LOGGER.debug('receive message from mq: %s', body)

        try:
            data = event._loads(body)
        except:
            data = body
        try:
            def handle_ack(future):
                if future.exception():
                    channel.basic_nack(deliver.delivery_tag)
                else:
                    channel.basic_ack(deliver.delivery_tag)

            def handle_timeout(future):
                if not future.done():
                    future.set_exception(Exception('timeout'))

            fu = util.run_async(self._handler, data)
            fu.add_done_callback(handle_ack)
            self._io_loop.call_later(settings.get('EVENT_HANDLE_TIMEOUT'), handle_timeout, fu)
        except:
            LOGGER.error('handle(%s) message error: %s', self._handler, traceback.format_exc())

    async def __connect(self):
        if self._connection and self._connection.is_open and not self._connection.is_closed:
            LOGGER.debug('open channel: %s', self._queue)
            channel = await _get_channel(self._connection)
            await gen.Task(channel.queue_declare, queue=self._queue, durable=True)
            await gen.Task(channel.queue_bind, queue=self._queue, exchange=self._exchange,
                           routing_key=self._routing_key)
            channel.basic_consume(self.__on_message, self._queue)
            channel.add_on_close_callback(self._on_close)
            self.running = True

    def _on_close(self, channel, reply_code, reply_text):
        LOGGER.debug('channel(%s) closed: (%s) %s', self._queue, reply_code, reply_text)
        self.running = False
        self._io_loop.call_later(1, self.__connect)

    def start(self, connection):
        if self.running:
            return
        self._connection = connection
        self._io_loop.spawn_callback(self.__connect)


class MqServer:
    def __init__(self, url, exchange, reconnect_delay=5.0, io_loop=None):
        self._url = url
        self._exchange = exchange
        self._reconnect_delay = reconnect_delay
        self._io_loop = io_loop or registry.io_loop()
        self._connection = None
        self._closing = False
        self._consumers = set()

    def _connect(self):
        self._connection = adapters.TornadoConnection(parameters=pika.URLParameters(self._url),
                                                      on_open_callback=self.__on_connection_open,
                                                      on_open_error_callback=self.__on_connection_open_error,
                                                      on_close_callback=self.__on_connection_closed)

    def __on_connection_open(self, connection):
        self._io_loop.spawn_callback(self.__perform_connection_open)

    async def __perform_connection_open(self):
        channel = await _get_channel(self._connection)
        await gen.Task(channel.exchange_declare, exchange=self._exchange, durable=True)
        for consumer in list(self._consumers):
            consumer.start(self._connection)
        channel.close()

    def __on_connection_open_error(self, connection, error):
        if self._closing:
            return
        LOGGER.warning('Connection Open failed, reopening in %s seconds: %s',
                       self._reconnect_delay, error)

        self._connection.add_timeout(self._reconnect_delay, self._connect)

    def __on_connection_closed(self, connection, reply_code, reply_text):
        if self._closing:
            return
        LOGGER.warning('Connection Closed, reopening in %s seconds: (%s) %s',
                       self._reconnect_delay, reply_code, reply_text)

        self._connection.add_timeout(self._reconnect_delay, self._connect)

    def add_consumer(self, queue, routing_key, handler):
        consumer = Consumer(self._exchange, queue, routing_key, handler, self._io_loop)
        consumer.start(self._connection)
        self._consumers.add(consumer)
        LOGGER.debug('add consumer(queue: %s, routing_key: %s, handler: %s) to %s',
                     queue, routing_key, util.meth_str(handler), self._url)

    async def publish_async(self, routing_key, data):
        try:
            channel = await _get_channel(self._connection)
            channel.basic_publish(self._exchange, routing_key, data.encode('utf-8'))
            channel.close()
            LOGGER.debug('success publish message to (exchange: %s, routing_key: %s), and data is: %s',
                         self._exchange, routing_key, data)
            return True
        except:
            LOGGER.error('fail publish message to (exchange: %s, routing_key: %s), and data is: %s',
                         self._exchange, routing_key, data)
            return False

    def start(self):
        self._closing = False
        self._connect()

    def stop(self):
        self._closing = True
        self._connection.close()



# coding: utf-8
from event_mq import event, mq_server, settings, registry

# register mq_servers
for name, data in settings.get('MQ_SERVERS').items():
    server = mq_server.MqServer(data['url'], data['exchange'],
                                data.get('reconnect_delay', 5),
                                registry.io_loop())
    server.start()
    registry.Registry._add_mq_server(name, server)



# coding: utf-8
from functools import wraps

from event_mq import event


def event_handler(event_name, mq_server='default'):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        event.subscribe(event_name, func, mq_server)
        return wrapper

    return decorator



# coding: utf-8
try:
    from user_settings import SETTINGS as user_settings
except ImportError:
    user_settings = {}

DEFAULTS = {
    'MQ_SERVERS': {
        'default': {
            'url': 'amqp://guest:guest@127.0.0.1:5672',
            'exchange': '',
            'reconnect_delay': 5.0
        }
    },
    'EVENT_HANDLE_TIMEOUT': 10,
    'EXECUTOR_MAX_WORKERS': 20
}


def get(attr):
    if attr not in DEFAULTS:
        raise AttributeError("Invalid setting: '%s'" % attr)

    try:
        # Check if present in user settings
        val = user_settings[attr]
    except KeyError:
        # Fall back to defaults
        val = DEFAULTS[attr]

    return val



# -*- coding:utf-8 -*-
from distutils.core import setup

setup(
    name='event-mq',
    version='0.2',
    packages=['event_mq'],
    description='A Event Queue for communication between progress.',
    url='https://github.com/songtao-git/event_mq',
    author='songtao',
    author_email='975765671@qq.com',
    license='MIT',
    keywords=['Event', 'progress communication'],
    install_requires=[
        'pika==0.11.0',
        'tornado==4.5.2',
    ],
)
